//! The runtime for executing oyster programs.
//! Panics when an invalid ast gets passed.

pub mod builtins;
mod pipeline;

use std::{
    ffi::OsString,
    fs::File,
    io::{self, Read},
    os::unix::{ffi::OsStringExt, io::FromRawFd},
    process,
};

use builtins::BuiltinMap;
use nix::{
    fcntl::OFlag,
    libc,
    sys::{
        signal::{self, SaFlags, SigAction, SigHandler, Signal},
        signalfd::SigSet,
        wait::{self, WaitPidFlag},
    },
    unistd::{self, ForkResult},
};
use oyster_parser::ast;
use thiserror::Error;

/// Exit status of a finished command.
#[derive(Debug)]
pub struct Status(pub i32);

impl Status {
    pub const SUCCESS: Status = Status(0);
    pub const COULD_NOT_EXEC: Status = Status(126);
    pub const COMMAND_NOT_FOUND: Status = Status(127);
    pub const SIG_BASE: Status = Status(128);
}

/// Errors that occur during runtime.
#[derive(Debug, Error)]
pub enum RuntimeError {
    #[error("failed to create pipe: {0}")]
    PipeCreationFailed(#[source] nix::Error),

    #[error("failed to dup fd: {0}")]
    DupFailed(#[source] nix::Error),

    #[error("command not found: {0:?}")]
    CommandNotFound(OsString),

    #[error("permission denied: {0:?}")]
    PermissionDenied(OsString),

    #[error("failed to spawn process: {0}")]
    SpawnFailed(#[source] io::Error),

    #[error("failed to fork: {0}")]
    ForkFailed(#[source] nix::Error),

    #[error("waitpid error: {0}")]
    WaitPid(nix::Error),

    #[error("I/O error: {0}")]
    IOError(#[source] io::Error),
}

pub struct Shell {
    is_running: bool,
    builtins: BuiltinMap,
}

impl Shell {
    /// Create a new shell.
    /// This also sets up the *global* signal handlers.
    pub fn new() -> Result<Shell, nix::Error> {
        let ignore = SigAction::new(SigHandler::SigIgn, SaFlags::empty(), SigSet::empty());

        unsafe {
            signal::sigaction(Signal::SIGTSTP, &ignore)?;
            signal::sigaction(Signal::SIGTTOU, &ignore)?;
        }

        Ok(Shell {
            is_running: true,
            builtins: Default::default(),
        })
    }

    /// Returns true if the shell is supposed to still be running.
    pub fn is_running(&self) -> bool {
        self.is_running
    }

    /// Runs a given program.
    pub fn run<'a>(&mut self, code: &'a ast::Code) -> Result<Status, RuntimeError> {
        let mut last_status = Status::SUCCESS;

        for stmt in code.0.iter() {
            last_status = match stmt {
                ast::Statement::Pipeline(pipeline) => self.run_pipeline(pipeline)?,
            };
        }

        Ok(last_status)
    }

    /// Runs a given program in a seperate process, and returns its output.
    pub fn run_subshell<'a>(&mut self, code: &'a ast::Code) -> Result<OsString, RuntimeError> {
        let (output, input) =
            unistd::pipe2(OFlag::empty()).map_err(RuntimeError::PipeCreationFailed)?;

        match unsafe { unistd::fork() }.map_err(RuntimeError::ForkFailed)? {
            ForkResult::Parent { child } => {
                let _ = unistd::close(input);
                let mut f = unsafe { File::from_raw_fd(output) };

                let mut bytes = Vec::new();
                f.read_to_end(&mut bytes).map_err(RuntimeError::IOError)?;

                let bytes = match bytes.iter().position(|b| !b.is_ascii_whitespace()) {
                    Some(start) => {
                        let end = bytes
                            .iter()
                            .rposition(|b| !b.is_ascii_whitespace())
                            .unwrap();

                        OsString::from_vec(bytes[start..=end].to_vec())
                    }
                    None => OsString::new(),
                };

                wait::waitpid(child, Some(WaitPidFlag::empty())).map_err(RuntimeError::WaitPid)?;

                Ok(bytes)
            }
            ForkResult::Child => {
                let _ = unistd::dup2(input, libc::STDOUT_FILENO);
                let _ = unistd::close(input);
                let _ = unistd::close(output);
                if let Err(err) = self.run(code) {
                    eprintln!("oyster: {}", err);
                }

                process::exit(0);
            }
        }
    }
}
